# Version file autogenerated by the build process
version = '0.13.5'
build = 1275
vcs_info = 'Git SHA1 acaf77d3 committed on 2023-10-20'
package_name = 'ksconf'

SHELL_VER_VARS = 'KSCONF_VERSION="0.13.5"\nKSCONF_BUILD="1275"\nKSCONF_VCS_INFO="Git SHA1 acaf77d3 committed on 2023-10-20"'

if __name__ == '__main__':
    print("# Please use 'python -m ksconf.version' instead.  This will eventually go away.")
    print(SHELL_VER_VARS)
